#!/usr/bin/env python
# -*- coding: utf-8 -*-
##
# motop - Unix "top" Clone for MongoDB
#
# Copyright (c) 2012, Tart İnternet Teknolojileri Ticaret AŞ
#
# Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby
# granted, provided that the above copyright notice and this permission notice appear in all copies.
#
# The software is provided "as is" and the author disclaims all warranties with regard to the software including all
# implied warranties of merchantability and fitness. In no event shall the author be liable for any special, direct,
# indirect, or consequential damages or any damages whatsoever resulting from loss of use, data or profits, whether
# in an action of contract, negligence or other tortious action, arising out of or in connection with the use or
# performance of this software.
##

"""Imports for Python 3 compatibility"""
from __future__ import print_function

"""Library imports"""
import json
from bson import json_util

"""Class imports"""
from .console import Block

class StatusBlock(Block):
    columnHeaders = ('Server', 'QPS', 'Active', 'Queue', 'Flush', 'Connection', 'Network I/O', 'Memory', 'Page Faults',)

    def __init__(self, servers):
        Block.__init__(self, self.columnHeaders)
        self.__servers = servers
        self.__oldStatus = {}

    def reset(self):
        lines = []

        for server in self.__servers:
            cells = []
            cells.append(server)
            status = server.status()
            if status:
                oldStatus = self.__oldStatus[server] if server in self.__oldStatus else status
                sec = status.deepgetDiff(oldStatus, 'uptimeMillis') / 1000.0
                if not sec:
                    sec = 1

                operations = sum(status.deepgetDiff(oldStatus, 'opcounters', k) for k in status.deepget('opcounters'))
                connectionsCurrent = status.deepget('connections', 'current') or 0
                connectionsAvailable = status.deepget('connections', 'available') or 0

                cells.append(operations / sec)
                cells.append(status.deepget('globalLock', 'activeClients', 'total'))
                cells.append(status.deepget('globalLock', 'currentQueue', 'total'))
                cells.append(status.deepgetDiff(oldStatus, 'backgroundFlushing', 'flushes') / sec)
                cells.append([connectionsCurrent, connectionsCurrent + connectionsAvailable])
                cells.append(status.deepget('network', ('bytesIn', 'bytesOut')))
                cells.append([v * 10**6 for v in status.deepget('mem', ('resident', 'mapped')) if v is not None])
                cells.append(status.deepgetDiff(oldStatus, 'extra_info', 'page_faults') / sec)

                self.__oldStatus[server] = status
            else:
                cells.append(server.lastError())

            lines.append(cells)

        Block.reset(self, lines)

class ServerBasedBlock(Block):
    def __init__(self, servers):
        Block.__init__(self, self.columnHeaders)
        self.__servers = servers
        self.__hiddenServers = []

    def findServer(self, name):
        for server in self.__servers:
            if server.sameServer(name):
                return server

    def connectedServers(self):
        return [server for server in self.__servers if server.connected() and server not in self.__hiddenServers]

    def hideServer(self, server):
        self.__hiddenServers.append(server)

class ReplicationInfoBlock(ServerBasedBlock):
    columnHeaders = ['Server', 'Source', 'SyncedTo', 'Inc']

    def reset(self):
        lines = []

        for server in self.connectedServers():
            replicationInfo = server.replicationInfo()
            if replicationInfo:
                cells = []
                cells.append(server)
                source = self.findServer(replicationInfo.get('host')) or replicationInfo.get('host')
                cells.append([replicationInfo.get('source'), source])
                if replicationInfo.get('syncedTo'):
                    cells.append(replicationInfo['syncedTo'].as_datetime())
                    cells.append(replicationInfo['syncedTo'].inc)

                lines.append(cells)
            else:
                self.hideServer(server)

        Block.reset(self, lines)

class ReplicaSetMemberBlock(ServerBasedBlock):
    columnHeaders = ('Server', 'Set', 'State', 'Uptime', 'Lag', 'Inc', 'Ping')

    def __add(self, line):
        """Merge same lines by revising the existent one."""
        for existentLine in self.__lines:
            if existentLine['set'] == line['set'] and existentLine['name'] == line['name']:
                for key in line.keys():
                    if line[key]:
                        if existentLine[key] < line[key]:
                            existentLine[key] = line[key]
                return

        self.__lines.append(line)

    def reset(self):
        self.__lines = []

        for server in self.connectedServers():
            replicaSetMembers = server.replicaSetMembers()
            if replicaSetMembers:
                for member in replicaSetMembers:
                    cells = []
                    cells.append(self.findServer(member.get('name')) or member.get('name'))
                    cells.append(member.get('set'))
                    cells.append(member.get('stateStr'))
                    cells.append(member.get('uptime'))
                    cells.append(member.get('pingMs'))
                    cells.append(member['date'] - member['optimeDate'] if 'optimeDate' in member else None)
                    cells.append(member.get('optime'))

                    self.__lines.append(cells)
            else:
                self.hideServer(server)

        Block.reset(self, self.__lines)

class Query:
    def __init__(self, **parts):
        """Translate query parts to arguments of pymongo find method."""
        self.__parts = {}

        if any([key in ('query', '$query') for key in parts.keys()]):
            for key, value in parts.items():
                if key[0] == '$':
                    key = key[1:]
                if key == 'query':
                    key = 'spec'
                if key == 'orderby':
                    key = 'sort'
                    value = list(value.items())
                if key == 'explain':
                    self.__explain = True

                self.__parts[key] = value
        else:
            self.__parts['spec'] = parts

    def __str__(self):
        return json.dumps(self.__parts, default=json_util.default)

    def print(self):
        """Print formatted query parts."""
        for key, value in self.__parts.items():
            print(key.title() + ':', end = ' ')
            if isinstance(value, list):
                print(', '.join([pair[0] + ': ' + str(pair[1]) for pair in value]))
            elif isinstance(value, dict):
                print(json.dumps(value, default=json_util.default, indent=4))
            else:
                print(value)

    def printExplain(self, server, ns):
        """Print the output of the explain command executed on the server."""
        explainOutput = server.explainQuery(ns, self.__parts)
        if not explainOutput:
            return False

        print()
        print('Cursor:', explainOutput['cursor'])
        print('Indexes:', end=' ')
        for index in explainOutput['indexBounds']:
            print(index, end=' ')
        print()
        print('IndexOnly:', explainOutput['indexOnly'])
        print('MultiKey:', explainOutput['isMultiKey'])
        print('Miliseconds:', explainOutput['millis'])
        print('Documents:', explainOutput['n'])
        print('ChunkSkips:', explainOutput['nChunkSkips'])
        print('Yields:', explainOutput['nYields'])
        print('Scanned:', explainOutput['nscanned'])
        print('ScannedObjects:', explainOutput['nscannedObjects'])
        if 'scanAndOrder' in explainOutput:
            print('ScanAndOrder:', explainOutput['scanAndOrder'])

        return True

class OperationBlock(Block):
    columnHeaders = ('Server', 'Opid', 'Client', 'Type', 'Sec', 'Locks', 'Namespace', 'Query')

    def __init__(self, servers, replicationOperationServers):
        Block.__init__(self, self.columnHeaders)
        self.__servers = servers
        self.__replicationOperationServers = replicationOperationServers

    def reset(self):
        self.__lines = []

        for server in self.__servers:
            if server.connected():
                hideReplicationOperations = server not in self.__replicationOperationServers
                for op in server.currentOperations(hideReplicationOperations):
                    cells = []
                    cells.append(server)
                    cells.append(str(op.get('opid')))
                    cells.append(op.get('client'))
                    cells.append(op.get('op'))
                    cells.append(op.get('secs_running'))

                    locks = []
                    if op.get('waitingForLock'):
                        locks.append('waiting')
                    if 'locks' in op:
                        if '^' in op['locks']:
                            """Do not show others if global lock exist."""
                            locks.append(op['locks']['^'])
                        else:
                            for ns, lock in op['locks'].items():
                                locks.append(lock + ' on ' + ns[1:])
                    elif 'lockType' in op:
                        locks.append(op['lockType'])

                    cells.append(locks)
                    cells.append(op.get('ns'))

                    if 'query' in op:
                        if '$msg' in op['query']:
                            cells.append(op['query']['$msg'])
                        else:
                            cells.append(Query(**op['query']))

                    self.__lines.append(cells)

        def sortKey(line): return line[4] or -1
        self.__lines.sort(key=sortKey, reverse=True)
        Block.reset(self, self.__lines)

    def __findServer(self, serverName):
        for server in self.__servers:
            if str(server) == serverName:
                return server

    def __findLine(self, serverName, opid):
        for line in self.__lines:
            if str(line[0]) == serverName and str(line[1]) == opid:
                return line

    def explainQuery(self, *parameters):
        line = self.__findLine(*parameters)
        if line and len(line) > 6 and line[6] and isinstance(line[7], Query):
            query = line[7]
            query.print()
            return query.printExplain(line[0], line[6])

    def kill(self, serverName, opid):
        server = self.__findServer(serverName)
        return server.killOperation(opid)

    def batchKill(self, second):
        """Kill operations running more than given seconds from top to bottom."""
        second = int(second)
        for line in self.__lines:
            if line[4] < second:
                """Do not look further as the list is reverse ordered by seconds."""
                break
            server = line[0]
            server.killOperation(line[1])

class QueryScreen:
    def __init__(self, console, chosenServers, autoKillSeconds=None):
        self.__console = console
        self.__servers = set(server for servers in chosenServers.values() for server in servers)

        self.__blocks = []
        self.__blocks.append(StatusBlock(chosenServers['status']))
        self.__blocks.append(ReplicationInfoBlock(chosenServers['replicationInfo']))
        self.__blocks.append(ReplicaSetMemberBlock(chosenServers['replicaSet']))
        self.__operationBlock = OperationBlock(chosenServers['operations'], chosenServers['replicationOperations'])
        self.__blocks.append(self.__operationBlock)

        self.__autoKillSeconds = autoKillSeconds

    def action(self):
        """Reset the blocks, refresh the console, perform actions for the pressed button."""
        button = None
        counter = 0

        while button != 'q':
            for block in self.__blocks:
                block.reset()
            self.__console.refresh(self.__blocks)
            button = self.__console.checkButton(1)
            counter += 1

            "Pause:"
            if button == 'p':
                button = self.__console.waitButton()

            "Single operation actions:"
            if button in ('e', 'k'):
                inputValues = self.__console.askForInput('Server', 'Opid')
                if inputValues:
                    if len(inputValues) == 2:
                        if button == 'e':
                            if not self.__operationBlock.explainQuery(*inputValues):
                                print('Explain failed. Try again.')
                        elif button == 'k':
                            if not self.__operationBlock.kill(*inputValues):
                                print('Kill failed. Try again.')
                    button = self.__console.waitButton()

            "Batch kill actions:"
            if button == 'K':
                inputValues = self.__console.askForInput('Sec')
                if inputValues:
                    self.__operationBlock.batchKill(*inputValues)
            if self.__autoKillSeconds is not None:
                self.__operationBlock.batchKill(self.__autoKillSeconds)

            "Reconnect actions:"
            if button in ('r', 'R') or counter % 20 == 0:
                for server in self.__servers:
                    if button == 'R' or not server.connected():
                        server.tryToConnect()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
##
# motop - Unix "top" Clone for MongoDB
#
# Copyright (c) 2012, Tart İnternet Teknolojileri Ticaret AŞ
#
# Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby
# granted, provided that the above copyright notice and this permission notice appear in all copies.
#
# The software is provided "as is" and the author disclaims all warranties with regard to the software including all
# implied warranties of merchantability and fitness. In no event shall the author be liable for any special, direct,
# indirect, or consequential damages or any damages whatsoever resulting from loss of use, data or profits, whether
# in an action of contract, negligence or other tortious action, arising out of or in connection with the use or
# performance of this software.
##

"""Imports for Python 3 compatibility"""
from __future__ import print_function
try:
    import __builtin__
    __builtin__.input = __builtin__.raw_input
except ImportError: pass

"""Library imports"""
import sys
import os
import tty
import termios
import struct
import fcntl
import select
import signal
import time
import numbers
from datetime import datetime

class Console:
    """Main class for input and output. Used with "with" statement to hide pressed buttons on the console."""
    def __init__(self):
        self.__deactiveConsole = DeactiveConsole(self)
        self.__saveSize()
        signal.signal(signal.SIGWINCH, self.__saveSize)
        self.__lastCheckTime = None

    def __enter__(self):
        """Hide pressed buttons on the console."""
        try:
            self.__settings = termios.tcgetattr(sys.stdin)
            tty.setcbreak(sys.stdin.fileno())
        except termios.error:
            self.__settings = None
        return self

    def __exit__(self, *ignored):
        if self.__settings:
            termios.tcsetattr(sys.stdin, termios.TCSADRAIN, self.__settings)

    def __saveSize(self, *ignored):
        try:
            self.__height, self.__width = struct.unpack('hhhh', fcntl.ioctl(0, termios.TIOCGWINSZ , '\000' * 8))[:2]
        except IOError:
            self.__height, self.__width = 20, 80

    def waitButton(self):
        while True:
            try:
                return sys.stdin.read(1)
            except IOError: pass

    def checkButton(self, waitTime):
        """Check one character input. Waits for approximately waitTime parameter as seconds."""
        if self.__lastCheckTime:
            timedelta = datetime.now() - self.__lastCheckTime
            waitTime -= timedelta.seconds + (timedelta.microseconds / 1000000.0)
        while waitTime > 0 and not select.select([sys.stdin], [], [], 0)[0]:
            time.sleep(0.1)
            waitTime -= 0.1
        self.__lastCheckTime = datetime.now()

        if select.select([sys.stdin], [], [], 0)[0]:
            return sys.stdin.read(1)

    def refresh(self, blocks):
        """Print the blocks with height and width left on the screen."""
        os.system('clear')
        leftHeight = self.__height
        for block in blocks:
            if not len(block):
                """Do not show the block if there are no lines."""
                continue
            if leftHeight <= 2:
                """Do not show the block if there are not enough lines left for header and a row."""
                break
            height = len(block) + 2 if len(block) + 2 < leftHeight else leftHeight

            try:
                block.print(height, self.__width)
                leftHeight -= height
                if leftHeight >= 2:
                    print()
                    leftHeight -= 1
            except IOError: pass

    def askForInput(self, *attributes):
        """Ask for input for given attributes in given order."""
        with self.__deactiveConsole:
            print()
            values = []
            for attribute in attributes:
                value = input(attribute + ': ')
                if not value:
                    break
                values.append(value)
            return values

class DeactiveConsole:
    """Class to use with "with" statement as "wihout" statement for Console class defined below."""
    def __init__(self, console):
        self.__console = console

    def __enter__(self):
        self.__console.__exit__()

    def __exit__(self, *ignored):
        self.__console.__enter__()

class Block:
    """Class to print blocks of ordered printables."""
    def __init__(self, columnHeaders):
        self.__columnHeaders = columnHeaders
        self.__columnWidths = [6] * len(self.__columnHeaders)

    def reset(self, lines):
        self.__lines = lines

    def __len__(self):
        return len(self.__lines)

    fixes = 'k', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y'

    def __cell(self, value):
        if isinstance(value, list):
            return ' / '.join(self.__cell(v) for v in value)

        if isinstance(value, numbers.Number):
            """Extents int to show big numbers human readable."""
            for fix in ('',) + self.fixes:
                if value < 10000:
                    return str(int(value)) + fix
                value = round(value / 1000)

        if value is not None:
            return str(value)

        return ''

    def __printLine(self, line, leftWidth, bold=False):
        """Print the cells separated by 2 spaces, cut the part after the width."""
        for index, value in enumerate(line):
            cell = self.__cell(value)
            if leftWidth < len(self.__columnHeaders[index]):
                """Do not show the column if there is not enough space for the header."""
                break
            if index + 1 < len(line):
                """Check the cell lenght if it is not the cell in the column. Set the column width to the cell lenght
                plus 2 for space if it is longer than the exisent column width."""
                self.__columnWidths[index] = max(len(cell) + 2, self.__columnWidths[index])

            if bold and sys.stdout.isatty():
                print('\x1b[1m', end='')
            print(cell.ljust(self.__columnWidths[index])[:leftWidth], end = '')
            if bold and sys.stdout.isatty():
                print('\x1b[0m', end='')
            leftWidth -= self.__columnWidths[index]

        """Finally, the new line."""
        print()

    def print(self, height, width):
        """Print the lines, cut the ones after the height."""
        assert height > 1
        self.__printLine(self.__columnHeaders, width, True)
        height -= 1
        for line in self.__lines:
            if height <= 1:
                break
            assert len(line) <= len(self.__columnHeaders)
            height -= 1
            self.__printLine(line, width)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
##
# motop - Unix "top" Clone for MongoDB
#
# Copyright (c) 2012, Tart İnternet Teknolojileri Ticaret AŞ
#
# Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby
# granted, provided that the above copyright notice and this permission notice appear in all copies.
#
# The software is provided "as is" and the author disclaims all warranties with regard to the software including all
# implied warranties of merchantability and fitness. In no event shall the author be liable for any special, direct,
# indirect, or consequential damages or any damages whatsoever resulting from loss of use, data or profits, whether
# in an action of contract, negligence or other tortious action, arising out of or in connection with the use or
# performance of this software.
##

"""Class imports"""
from libmotop.console import Console
from libmotop.server import Server
from libmotop.queryscreen import QueryScreen

"""Two attempts to import the same class for Python 3 compatibility."""
try:
    from ConfigParser import SafeConfigParser
except ImportError:
    from configparser import SafeConfigParser

"""Metadata"""
from libmotop import __name__, __version__, __doc__

"""Main configuration"""
configFile = '/etc/motop.conf'
optionalVariables = ('username', 'password')
choices = ('status', 'replicationInfo', 'replicaSet', 'operations', 'replicationOperations')

def version():
    return __name__ + ' ' + str(__version__)

def parseArguments():
    """Create ArgumentParser instance. Return parsed arguments."""
    from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter, description=__doc__)
    parser.add_argument('hosts', metavar='host', nargs='*', default=('localhost:27017',),
            help='address of the server or section name on the configuration file')
    parser.add_argument('-u', '--username', dest='username', help='username for authentication')
    parser.add_argument('-p', '--password', dest='password', help='password for authentication')
    parser.add_argument('-c', '--conf', dest='conf', default=configFile,
            help='path of configuration file')
    parser.add_argument('-V', '--version', action='version', version=version())
    parser.add_argument('-K', '--auto-kill', dest='autoKillSeconds',
            help='seconds to kill operations automatically')
    return parser.parse_args()

def commonServers(config, arguments):
    """First try to match servers on the config with the ones on the arguments."""
    servers = []
    for host in arguments.hosts:
        for section in config.sections():
            if section == host:
                servers.append(Server(section, **dict(config.items(section))))
    if servers:
        return servers

    """Second use the servers on the config."""
    if config.sections():
        return [Server(section, **dict(config.items(section))) for section in config.sections()]

    """Third use the servers on the arguments."""
    return [Server(host, host, arguments.username, arguments.password) for host in arguments.hosts]

def run():
    """Get the arguments and parse the config file. Activate console. Get servers from the config file
    or from arguments. Show the query screen."""
    arguments = parseArguments()
    config = SafeConfigParser({'username': arguments.username, 'password': arguments.password})
    config.read(arguments.conf)
    servers = commonServers(config, arguments)

    chosenServers = {}
    for choice in choices:
        if config.sections():
            chosenServers[choice] = []
            for server in servers:
                if not config.has_option(str(server), choice) or config.getboolean(str(server), choice):
                    chosenServers[choice].append(server)
        else:
            chosenServers[choice] = servers

    with Console() as console:
        queryScreen = QueryScreen(console, chosenServers, autoKillSeconds=arguments.autoKillSeconds)
        try:
            queryScreen.action()
        except KeyboardInterrupt: pass




#!/usr/bin/env python

from setuptools import setup
import libmotop

def readme():
    with open('README.rst') as readmeFile:
        return readmeFile.read()

setup(name=libmotop.__name__,
        version=str(libmotop.__version__),
        packages=('libmotop',),
        scripts=('motop',),
        install_requires=('pymongo', 'argparse'),
        author='Emre Hasegeli',
        author_email='hasegeli@tart.com.tr',
        license='ICS',
        url='https://github.com/tart/motop',
        platforms='POSIX',
        description=libmotop.__doc__,
        keywords='mongo realtime monitoring examine explain kill operations',
        long_description=readme())




# -*- coding: utf-8 -*-
##
# motop - Unix "top" Clone for MongoDB
#
# Copyright (c) 2012, Tart İnternet Teknolojileri Ticaret AŞ
#
# Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby
# granted, provided that the above copyright notice and this permission notice appear in all copies.
#
# The software is provided "as is" and the author disclaims all warranties with regard to the software including all
# implied warranties of merchantability and fitness. In no event shall the author be liable for any special, direct,
# indirect, or consequential damages or any damages whatsoever resulting from loss of use, data or profits, whether
# in an action of contract, negligence or other tortious action, arising out of or in connection with the use or
# performance of this software.
##

__name__ = 'motop'
__version__ = 4.2
__doc__ = '"Top" clone for MongoDB.'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
##
# motop - Unix "top" Clone for MongoDB
#
# Copyright (c) 2012, Tart İnternet Teknolojileri Ticaret AŞ
#
# Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby
# granted, provided that the above copyright notice and this permission notice appear in all copies.
# 
# The software is provided "as is" and the author disclaims all warranties with regard to the software including all
# implied warranties of merchantability and fitness. In no event shall the author be liable for any special, direct,
# indirect, or consequential damages or any damages whatsoever resulting from loss of use, data or profits, whether
# in an action of contract, negligence or other tortious action, arising out of or in connection with the use or
# performance of this software.
##

"""Library imports"""
import os
import sys
import time
import pymongo

class Server:
    def __init__(self, name, address, username=None, password=None):
        self.__name = name
        self.__address = address
        self.__username = username
        self.__password = password
        self.tryToConnect()

    connectionClass = pymongo.MongoClient if pymongo.version_tuple >= (2, 4) else pymongo.Connection
    connectionParemeters = {'connectTimeoutMS': 1000, 'read_preference': pymongo.ReadPreference.SECONDARY}

    def tryToConnect(self):
        self.__connection = None
        try:
            self.__connection = self.connectionClass(self.__address, **self.connectionParemeters)
        except pymongo.errors.ConnectionFailure as error:
            self.__lastError = error

        if self.__username and self.__password:
            self.__connection.admin.authenticate(self.__username, self.__password)

    def __str__(self):
        return self.__name

    def sameServer(self, name):
        if self.__name == name:
            return True
        if self.__address == name:
            return True
        if ':' not in self.__address and self.__address + ':' + str(self.connectionClass.PORT) == name:
            return True
        return False

    def connected(self):
        return self.__connection is not None

    def __execute(self, procedure, *args, **kwargs):
        """Try 10 times to execute the procedure."""
        for tryCount in range(10):
            try:
                return procedure(*args, **kwargs)
            except pymongo.errors.AutoReconnect as error:
                self.__lastError = error
                time.sleep(0.1)
            except pymongo.errors.OperationFailure as error:
                self.__lastError = error
                break

    def __executeYield(self, *args, **kwargs):
        """Execute the procedure and yield items until get next item fails."""
        try:
            for item in self.__execute(*args, **kwargs):
                yield item
        except pymongo.errors.AutoReconnect as error:
            self.__lastError = error

    def lastError(self):
        return self.__lastError

    def status(self):
        if self.connected():
            result = self.__execute(self.__connection.admin.command, 'serverStatus')

            if result:
                return Result(result)

    def replicationInfo(self):
        """Find replication source from the local collection."""
        for source in self.__executeYield(self.__connection.local.sources.find):
            return Result(source)

    def replicaSetMembers(self):
        """Execute replSetGetStatus operation on the server. Filter arbiters. Calculate the lag. Add relation to the
        member which is the server itself. Return the replica set."""
        replicaSetStatus = self.__execute(self.__connection.admin.command, 'replSetGetStatus')
        if replicaSetStatus:
            for member in replicaSetStatus['members']:
                if member.get('statusStr') not in ['ARBITER']:
                    member['set'] = replicaSetStatus.get('set')
                    member['date'] = replicaSetStatus.get('date')

                    yield Result(member)

    def currentOperations(self, hideReplicationOperations=False):
        """Execute currentOp operation on the server. Filter and yield returning operations."""
        operations = self.__execute(self.__connection.admin.current_op)
        if operations:
            for op in operations['inprog']:
                if hideReplicationOperations:
                    if op.get('op') == 'getmore' and op.get('ns').startswith('local.oplog.'):
                        """Condition to find replication operation on the master."""
                        continue
                    if op.get('op') and op.get('ns') in ('', 'local.sources'):
                        """Condition to find replication operation on the slave."""
                        continue

                yield Result(op)

    def explainQuery(self, namespace, findParameters):
        databaseName, collectionName = namespace.split('.', 1)
        collection = getattr(getattr(self.__connection, databaseName), collectionName)
        cursor = self.__execute(collection.find, **findParameters)
        return self.__execute(cursor.explain)

    def killOperation(self, opid):
        """Kill operation using the "mongo" executable on the shell. That is because I could not make it with
        pymongo."""
        command = "echo 'db.killOp({0})' | mongo".format(str(opid))
        command += ' ' + self.__address + '/admin'
        if self.__username:
            command += ' --username ' + self.__username
        if self.__password:
            command += ' --password ' + self.__password

        exitCode = os.system(command)
        return exitCode == 0

class Result(dict):
    def deepget(self, arg, *args):
        if isinstance(arg, tuple):
            return [self.deepget(a, *args) for a in arg]

        if arg in self:
            if args:
                return Result(self[arg]).deepget(*args)
            return self[arg]

        return None

    def deepgetDiff(self, other, *args):
        value = self.deepget(*args)
        otherValue = other.deepget(*args)
        if value and otherValue:
            return value - otherValue
        return 0

